/*
 * Copyright (c) 2003-2007 Willem Dijkstra
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "error.h"
#include "drivers.h"
#include "driver_cf.h"

void
cf_connect(struct driver *d)
{
    struct termios tp;

    d->capabilities =  DC_BACKLIGHT | DC_CONTRAST;

    if ((d->control.cf.fd = open(d->connect, O_RDWR | O_NONBLOCK)) == -1) {
        fatal("%s: cannot open serial '%s': %.200s",
              d->driver, d->connect, strerror(errno));
    }

    if (tcgetattr(d->control.cf.fd, &tp) == -1) {
        fatal("%s: cannot get terminal attributes: %.200s",
              d->driver, strerror(errno));
    }

    if (cfsetospeed(&tp, d->speed) == -1) {
        fatal("%s: cannot set output speed: %.200s",
              d->driver, strerror(errno));
    }

    if (cfsetispeed(&tp, B0) == -1) {
        fatal("%s: cannot set input speed: %.200s",
              d->driver, strerror(errno));
    }

    tp.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
    tp.c_oflag &= ~OPOST;
    tp.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
    tp.c_cflag &= ~(CSIZE | PARENB | CRTSCTS);
    tp.c_cflag |= CS8 | CREAD | CLOCAL;
    tp.c_cc[VMIN] = 1;
    tp.c_cc[VTIME] = 3;

    if (tcsetattr(d->control.cf.fd, TCSANOW, &tp) == -1) {
        fatal("%s: cannot not set serial line %s to raw: %.200s",
              d->driver, d->connect, strerror(errno));
    }

    cf_setupblock(d);
    cf_setupvbar(d);
}
void
cf_disconnect(struct driver *d)
{
    if (cf_isconnected(d)) {
        close(d->control.cf.fd);
        bzero(d, sizeof(struct driver));
    }
}
int
cf_isconnected(struct driver *d)
{
    return (d->control.cf.fd != 0);
}
void
cf_clear(struct driver *d)
{
    // TODO - use scroll, line feed?
}
char
cf_gethchar(struct driver *d, int percentage)
{
    fatal("%s: horizontal bars not implemented", d->driver);

    return 0; /* NOT REACHED */
}
char *
cf_getoption()
{
    return "todo";
    // TODO
}
char
cf_getvchar(struct driver *d, int percentage)
{
    int unit = 1000 / 8;
    int i;

    percentage %= 101;
    percentage += 5;
    percentage *= 10;

    i = percentage / unit;

    if (i == 0) {
        return CF_CHAR_SPACE;
    }

    if (i == 8) {
        return CF_CHAR_BLOCK;
    }

    return CF_CHAR_VBAR + i - 1;
}
void
cf_flush(struct driver *d)
{
    fsync(d->control.cf.fd);
}
void
cf_nprintf(struct driver *d, int maxsize, char *fmt, ...)
{
    va_list args;
    int size;

    maxsize %= d->buflen;

    va_start(args, fmt);
    vsnprintf(d->buffer, maxsize, fmt, args);
    va_end(args);
    size = strlen(d->buffer);

    write(d->control.cf.fd, d->buffer, size);
}
void
cf_option(struct driver *d, int argc, char **argv)
{
    // TODO
}
void
cf_printf(struct driver *d, char *fmt, ...)
{
    va_list args;
    int size;

    va_start(args, fmt);
    size = vsnprintf(d->buffer, d->buflen, fmt, args);
    va_end(args);

    write(d->control.cf.fd, d->buffer, size);
}
void
cf_reset(struct driver *d)
{
    /* reset lcd into known state */
    cf_printf(d, "         %c%c", CFCMD_REBOOT, CFCMD_REBOOT);
    cf_flush(d);
    sleep(1);

    /* set our options */
    cf_setcontrast(d, CF_DEFAULT_CONTRAST);
    cf_setbacklight(d, CF_DEFAULT_BACKLIGHT);
    cf_flush(d);

    cf_printf(d, "%c", CFCMD_WRAP_OFF);
    cf_flush(d);
    cf_printf(d, "%c", CFCMD_SCROLL_OFF);
    cf_flush(d);
    cf_printf(d, "%c", CFCMD_FORM_FEED);
    cf_flush(d);
    cf_printf(d, "%c", CFCMD_HIDE_CURSOR);
    cf_flush(d);

    /* set our special chars */
    cf_setupblock(d);
    cf_setupvbar(d);

    d->x = d->y = 0;
}
void
cf_setbacklight(struct driver *d, int backlight)
{
    cf_printf(d, "%c%c",
              CFCMD_BACKLIGHT_CONTROL,
              backlight % 101);
}
void
cf_setcapabilities(struct driver *d, int capabilities)
{
    if (capabilities & DC_BACKLIGHT) {
        warning("%s: backlight support cannot be detected", d->driver);
        capabilities &= ~DC_BACKLIGHT;
    }

    if (capabilities & DC_BLOCK) {
        if (!(d->capabilities & DC_BLOCK)) {
            cf_setupblock(d);
        }
        capabilities &= ~DC_BLOCK;
    }

    if (capabilities & DC_CONTRAST) {
        capabilities &= ~DC_CONTRAST;
    }

    if (capabilities & DC_HBAR) {
        cf_setuphbar(d);
        capabilities &= ~DC_HBAR;
    }

    if (capabilities & DC_VBAR) {
        cf_setupvbar(d);
        capabilities &= ~DC_VBAR;
    }

    if (capabilities != 0) {
        fatal("%s: unknown capability 0x%08x requested", d->driver, capabilities);
    }
}
void
cf_setcontrast(struct driver *d, int contrast)
{
    cf_printf(d, "%c%c",
              CFCMD_CONTRAST_CONTROL,
              contrast % 101);
}
void
cf_setspecialchar(struct driver *d, int ch, struct specialchar *data)
{
    int ch_data[8];
    int i, bit, mask;

    if ((ch < CF_CHAR_SPECIAL) || (ch > CF_CHAR_SPECIAL + 7)) {
        fatal("%s: cannot set specialchar %d",
              d->driver, ch);
    }

    for (i = 0; i < 8; i ++) {
        ch_data[i] = 0;
        mask = 0xF;

        for (bit = 0; bit < 6; bit ++) {
            ch_data[i] += ((data->line[i] & mask) ? (1 << bit) : 0);
            mask <<= 4;
        }
    }

    cf_printf(d, "%c%c%c%c%c%c%c%c%c%c",
              CFCMD_SET_CUSTOM_CHARACTER_BITMAP,
              ch - CF_CHAR_SPECIAL,
              ch_data[0], ch_data[1], ch_data[2], ch_data[3],
              ch_data[4], ch_data[5], ch_data[6], ch_data[7]);
}
void
cf_setxy(struct driver *d, int x, int y)
{
    d->x = x % d->maxx;
    d->y = y % d->maxy;
    cf_printf(d, "%c%c%c",
              CFCMD_SET_CURSOR_POSITION,
              d->x, d->y);
}
void
cf_setupblock(struct driver *d)
{
    struct specialchar block = {
        {
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
        }
    };

    cf_setspecialchar(d, CF_CHAR_BLOCK, &block);
    d->capabilities |= DC_BLOCK;
}
void
cf_setuphbar(struct driver *d)
{
    /* only 7 custom chars allowed, we use these for vbars */
    fatal("%s: horizontal bar not supported in this driver", d->driver);
}
void
cf_setupvbar(struct driver *d)
{
    int i;
    struct specialchar vbar[] = {
        {{
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x111111,
        }},
        {{
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x111111,
            0x111111,
        }},
        {{
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x111111,
            0x111111,
            0x111111,
        }},
        {{
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
        }},
        {{
            0x000000,
            0x000000,
            0x000000,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
        }},
        {{
            0x000000,
            0x000000,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
        }},
        {{
            0x000000,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
        }}
    };

    for (i = 0; i < (sizeof(vbar) / sizeof(struct specialchar)); i++) {
        cf_setspecialchar(d, CF_CHAR_VBAR + i, &vbar[i]);
    }

    d->capabilities &= !DC_HBAR;
    d->capabilities |= DC_VBAR;
}
