/*
 * Copyright (c) 2007 Willem Dijkstra
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/select.h>

#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "error.h"
#include "drivers.h"
#include "driver_cfp.h"
#include "rbuffer.h"

__BEGIN_DECLS
void cfp_send_once(struct driver *d, uint8_t cmd, char *fmt, va_list args);
uint16_t cfp_crc16(struct cfpacket *p);
uint16_t cfp_crc16_start();
uint16_t cfp_crc16_update(uint16_t crc, uint8_t byte);
uint16_t cfp_crc16_update_block(uint16_t crc, const void *buf, uint32_t len);
uint16_t cfp_crc16_finish(uint16_t crc);
void cfp_crc16_init_table();
void cfp_dump(struct driver *d, struct cfpacket *p);
void cfp_send(struct driver *d, uint8_t cmd, char *fmt, ...);
__END_DECLS

uint16_t cfp_crc16_table[256];

void
cfp_connect(struct driver *d)
{
    struct termios tp;

    cfp_crc16_init_table();

    d->capabilities = DC_CONTRAST | DC_BACKLIGHT;

    if ((d->control.cfp.fd = open(d->connect, O_RDWR | O_NONBLOCK)) == -1) {
        fatal("%s: cannot open serial '%s': %.200s",
              d->driver, d->connect, strerror(errno));
    }

    if (tcgetattr(d->control.cfp.fd, &tp) == -1) {
        fatal("%s: cannot get terminal attributes: %.200s",
              d->driver, strerror(errno));
    }

    if (cfsetospeed(&tp, d->speed) == -1) {
        fatal("%s: cannot set output speed: %.200s",
              d->driver, strerror(errno));
    }

    if (cfsetispeed(&tp, d->speed) == -1) {
        fatal("%s: cannot set input speed: %.200s",
              d->driver, strerror(errno));
    }

    tp.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
    tp.c_oflag &= ~OPOST;
    tp.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
    tp.c_cflag &= ~(CSIZE | PARENB | CRTSCTS);
    tp.c_cflag |= CS8 | CREAD | CLOCAL;
    tp.c_cc[VMIN] = 1;
    tp.c_cc[VTIME] = 3;

    if (tcsetattr(d->control.cfp.fd, TCSANOW, &tp) == -1) {
        fatal("%s: cannot not set serial line %s to raw: %.200s",
              d->driver, d->connect, strerror(errno));
    }

    rbuffer_clear(&d->control.cfp.rb);
    rbuffer_attach(&d->control.cfp.rb, d->control.cfp.fd);

    cfp_setupblock(d);
    cfp_setupvbar(d);
}
void
cfp_disconnect(struct driver *d)
{
    if (cfp_isconnected(d)) {
        close(d->control.cfp.fd);
        bzero(d, sizeof(struct driver));
    }
}
int
cfp_isconnected(struct driver *d)
{
    return (d->control.cfp.fd != 0);
}
void
cfp_clear(struct driver *d)
{
    int lcdsize = d->maxx * d->maxy;

    if ((d->buffer != NULL) && (d->buflen > 0)) {
        snprintf(d->buffer, d->buflen, "xxxyyy%%% ds", lcdsize);
        cfp_setxy(d, 0, 0);
        cfp_printf(d, d->buffer, " ");
    }
}
char
cfp_gethchar(struct driver *d, int percentage)
{
    fatal("%s: horizontal bars not implemented", d->driver);

    return 0; /* NOT REACHED */
}
char
cfp_getvchar(struct driver *d, int percentage)
{
    int unit = 1000 / 7;
    int i;

    percentage %= 101;
    percentage += 5;
    percentage *= 10;

    i = percentage / unit;

    if (i == 0) {
        return CFP_CHAR_SPACE;
    }

    return CFP_CHAR_VBAR + i - 1;
}
void
cfp_flush(struct driver *d)
{
    fsync(d->control.cfp.fd);
}
void
cfp_nprintf(struct driver *d, int maxsize, char *fmt, ...)
{
    va_list args;
    int size;

    maxsize %= d->buflen;

    va_start(args, fmt);
    vsnprintf(d->buffer, maxsize, fmt, args);
    va_end(args);
    size = strlen(d->buffer);

    cfp_send(d, CFPCMD_PRINT, "ddn", d->x, d->y, size, d->buffer);
}
void
cfp_printf(struct driver *d, char *fmt, ...)
{
    va_list args;
    int size;

    va_start(args, fmt);
    size = vsnprintf(d->buffer, CFP_MAX_DATA_LENGTH - d->x, fmt, args);
    va_end(args);

    cfp_send(d, CFPCMD_PRINT, "ddn", d->x, d->y, size, d->buffer);
}
void
cfp_reset(struct driver *d)
{
    /* reset lcd into known state */
    cfp_send(d, CFPCMD_RESET, "ddd",
             CFPARG_RESET_LCD_0,
             CFPARG_RESET_LCD_1,
             CFPARG_RESET_LCD_2);

    sleep(1);
    /* set our options */
    cfp_setcontrast(d, CFP_DEFAULT_CONTRAST);
    cfp_setbacklight(d, CFP_DEFAULT_BACKLIGHT);

    cfp_send(d, CFPCMD_SET_CURSOR_STYLE, "d", 0);

    /* CFPCMD_CLEAR will fill the screen with the welcome screen, which may or
     * may not be the empty screen */
    cfp_send(d, CFPCMD_CLEAR, "");
    cfp_clear(d);

    /* set our special chars */
    cfp_setupblock(d);
    cfp_setupvbar(d);

    d->x = d->y = 0;
}
void
cfp_setbacklight(struct driver *d, int backlight)
{
    cfp_send(d, CFPCMD_SET_BACKLIGHT, "d", backlight);
}
void
cfp_setcapabilities(struct driver *d, int capabilities)
{
    if (capabilities & DC_BACKLIGHT) {
        warning("%s: backlight support cannot be detected", d->driver);
        capabilities &= ~DC_BACKLIGHT;
    }

    if (capabilities & DC_BLOCK) {
        if (!(d->capabilities & DC_BLOCK)) {
            cfp_setupblock(d);
        }
        capabilities &= ~DC_BLOCK;
    }

    if (capabilities & DC_CONTRAST) {
        capabilities &= ~DC_CONTRAST;
    }

    if (capabilities & DC_HBAR) {
        cfp_setuphbar(d);
        capabilities &= ~DC_HBAR;
    }

    if (capabilities & DC_VBAR) {
        cfp_setupvbar(d);
        capabilities &= ~DC_VBAR;
    }

    if (capabilities != 0) {
        fatal("%s: unknown capability 0x%08x requested", d->driver, capabilities);
    }
}
void
cfp_setcontrast(struct driver *d, int contrast)
{
    cfp_send(d, CFPCMD_SET_CONTRAST, "d", contrast);
}
void
cfp_setspecialchar(struct driver *d, int ch, struct specialchar *data)
{
    int ch_data[8];
    int i, bit, mask;

    if ((ch < CFP_CHAR_SPECIAL) || (ch > CFP_CHAR_SPECIAL + 7)) {
        fatal("%s: cannot set specialchar %d",
              d->driver, ch);
    }

    for (i = 0; i < 8; i ++) {
        ch_data[i] = 0;
        mask = 0xF;

        for (bit = 0; bit < 6; bit ++) {
            ch_data[i] += ((data->line[i] & mask) ? (1 << bit) : 0);
            mask <<= 4;
        }
    }

    cfp_send(d, CFPCMD_SET_SPECIAL_CHARACTER_DATA, "ddddddddd",
             ch - CFP_CHAR_SPECIAL,
             ch_data[0], ch_data[1], ch_data[2], ch_data[3],
             ch_data[4], ch_data[5], ch_data[6], ch_data[7]);
}
void
cfp_setupblock(struct driver *d)
{
    struct specialchar block = {
        {
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
        }
    };

    cfp_setspecialchar(d, CFP_CHAR_BLOCK, &block);
    d->capabilities |= DC_BLOCK;
}
void
cfp_setuphbar(struct driver *d)
{
    /* only 7 custom chars allowed, we use these for vbars */
    fatal("%s: horizontal bar not supported in this driver", d->driver);
}
void
cfp_setupvbar(struct driver *d)
{
    int i;
    struct specialchar vbar[] = {
        {{
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x111111,
            0x000000,
        }},
        {{
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x111111,
            0x111111,
            0x000000,
        }},
        {{
            0x000000,
            0x000000,
            0x000000,
            0x000000,
            0x111111,
            0x111111,
            0x111111,
            0x000000,
        }},
        {{
            0x000000,
            0x000000,
            0x000000,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x000000,
        }},
        {{
            0x000000,
            0x000000,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x000000,
        }},
        {{
            0x000000,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x000000,
        }},
        {{
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x111111,
            0x000000,
        }}
    };

    for (i = 0; i < (sizeof(vbar) / sizeof(struct specialchar)); i++) {
        cfp_setspecialchar(d, CFP_CHAR_VBAR + i, &vbar[i]);
    }

    d->capabilities &= !DC_HBAR;
    d->capabilities |= DC_VBAR;
}
void
cfp_setxy(struct driver *d, int x, int y)
{
    d->x = x % d->maxx;
    d->y = y % d->maxy;

    cfp_send(d, CFPCMD_SET_CURSOR_POSITION, "dd", d->x, d->y);
}
char *
cfp_getoption(struct driver *d)
{
/* TODO
Ping
Get hardware of firmware version
Read user flash
Write user flash
Store current state as boot state
Reset CFA
Reset Host (SCAB required)
Power off Host (SCAB required)
Clear LCD"
*/
    return "TODO";
}
void
cfp_option(struct driver *d, int argc, char **argv)
{
    // TODO
}

int
cfp_readpacket(struct driver *d)
{
    size_t len;
    struct timeval tv = {0, CFP_MAX_PACKET_WAIT};
    uint32_t i;
    uint8_t c;
    struct rbuffer *r = &d->control.cfp.rb;
    struct cfpacket *p = &d->control.cfp.p;

    len = rbuffer_len(r);
    if (len < sizeof(struct cfpacket)) {
        rbuffer_get_data_wait(r, &len, &tv);
    }

    if (rbuffer_get_byte(r, &p->command) == 0 ||
        rbuffer_get_byte(r, &p->length) == 0) {
        return 0;
    }

    if (p->length >= CFP_MAX_DATA_LENGTH) {
        /* maybe p->length was the start of a new packet
           unget the byte to allow resynchronisation */
        rbuffer_unget_byte(r, p->length);
        return 0;
    }

    len = rbuffer_len(r);

    if (len < p->length + 2) {
        /* should have had the entire packet in memory by now */
        rbuffer_unget_byte(r, p->length);
        return 0;
    }

    for (i = 0; i < p->length && rbuffer_get_byte(r, &p->data[i]) == 1; i++);

    if (i != p->length) {
        /* failure during getting the bytes */
        return 0;
    }

    if (rbuffer_get_byte(r, &c)) {
        p->crc = c;
        if (rbuffer_get_byte(r, &c)) {
            p->crc = (c << 8) | p->crc;

            if (cfp_crc16(p) == p->crc) {
                return 1;
            }
        }
    }

    return 0;
}
void
cfp_identify(struct driver *d)
{
    /* TODO Rename */
    cfp_send(d, CFPCMD_GET_VERSION, "");
}
uint16_t
cfp_crc16(struct cfpacket *p)
{
    uint32_t i;
    uint16_t crc;

    crc = cfp_crc16_start();
    crc = cfp_crc16_update(crc, p->command);
    crc = cfp_crc16_update(crc, p->length);
    for (i = 0; i < p->length; i++)
        crc = cfp_crc16_update(crc, p->data[i]);

    return cfp_crc16_finish(crc);
}
void
cfp_crc16_init_table()
{
    unsigned int i, j;
    uint16_t d;
    uint16_t c = 0;

    for (i = 0; i < 256; ++i) {
        c = 0;
        d = i;

        for (j = 8; j > 0; --j) {
                c = ((c ^ d) & 1) ? (c >> 1) ^ CFP_CRCPOLY : (c >> 1);
                d = d >> 1;
        }

        cfp_crc16_table[i] = c;
    }
}
uint16_t
cfp_crc16_finish(uint16_t crc)
{
    return ~crc;
}
uint16_t
cfp_crc16_start()
{
    return 0xffff;
}
uint16_t
cfp_crc16_update(uint16_t crc, uint8_t byte)
{
    return (crc >> 8) ^ cfp_crc16_table[(crc ^ byte) & 0xff];
}
uint16_t
cfp_crc16_update_block(uint16_t crc, const void *buf, uint32_t len)
{
    uint8_t *p;

    for (p = (u_int8_t *) buf; len > 0; ++p, --len)
        crc = (crc >> 8) ^ cfp_crc16_table[(crc ^ *p) & 0xff];

    return crc;

}
void
cfp_dump(struct driver *d, struct cfpacket *p)
{
    uint32_t len = 0;
    uint8_t t = (p->command & 0xc0) >> 6;
    char *types[4] = { "host->cfa", "cfa->host",
                       "cfa report->host", "cfa error->host" };

    len = snprintf(d->buffer, d->buflen,
                   "packet = [t = %s c = 0x%02x l = 0x%02x "
                   "crc = 0x%04x data =",
                   types[t], (p->command & 0x3f), p->length,
                   p->crc);

    for (t = 0; t < p->length && len < (d->buflen - 1); t++) {
        len += snprintf(d->buffer + len, d->buflen - len, " 0x%02x", p->data[t]);
    }

    len += snprintf(d->buffer + len, d->buflen - len, "]");

    d->buffer[d->buflen-1] = '\0';

    debug("%s: packet dump: %s", d->driver, d->buffer);
}
void
cfp_send(struct driver *d, uint8_t cmd, char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    cfp_send_once(d, cmd, fmt, ap);
    va_end(ap);

    while (cfp_readpacket(d)) {
        if ((d->control.cfp.p.command & 0x3f) == cmd) {
            return;
        } else {
            debug("expected 0x%02x", cmd);
            cfp_dump(d, &d->control.cfp.p);
        }
    }
}
void
cfp_send_once(struct driver *d, uint8_t cmd, char *fmt, va_list args)
{
    uint8_t raw[CFP_MAX_PACKET_LENGTH];
    uint32_t r = 0;
    uint32_t n;
    uint32_t f;
    uint32_t s;
    char *str;
    uint16_t crc = cfp_crc16_start();
    int32_t result;

    bzero(&raw, sizeof(raw));
    raw[r++] = cmd;

    /* packet length is determined after all args have been filled in */
    raw[r++] = 0;

    for (f = 0; r < CFP_MAX_PACKET_LENGTH - 2 && fmt[f] != '\0'; f++) {
        switch (fmt[f]) {
        case 'd':
            raw[r++] = va_arg(args, int);
            break;

        case 's':
            str = va_arg(args, char *);
            for (s = 0; r < CFP_MAX_PACKET_LENGTH - 2 &&
                     str[s] != '\0'; s++, r++)
                raw[r] = str[s];
            break;

        case 'n':
            n = va_arg(args, int);
            str = va_arg(args, char *);
            for (s = 0; r < CFP_MAX_PACKET_LENGTH - 2 &&
                     str[s] != '\0' &&
                     s < n; s++, r++)
                raw[r] = str[s];
            break;

        default:
            warning("unknown format character");
        }
    }
    /* set length */
    raw[1] = r - 2;

    crc = cfp_crc16_update_block(crc, &raw[0], r);
    crc = cfp_crc16_finish(crc);
    raw[r++] = crc & 0xff;
    raw[r++] = (crc >> 8) & 0xff;

    debug("sending cmd 0x%02x", cmd);
    if ((result = write(d->control.cfp.fd, &raw[0], r)) < r) {
        warning("could not write entire packet: %.200s", strerror(errno));
    }
    fsync(d->control.cfp.fd);
}
