/*
 * Copyright (c) 2005-2007 Willem Dijkstra
 *                         Jorge Sanchez Pelaez beto
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/types.h>

#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "error.h"
#include "drivers.h"
#include "driver_hd44780.h"
#include "port.h"
#include "time.h"

__BEGIN_DECLS
void hd44780_writecontrol(struct driver *d, int);
void hd44780_writedata(struct driver *d, int);
__END_DECLS

char hd44780_rowstart[4] = HD_ROWSTART;

void
hd44780_connect(struct driver *d)
{
    long portnum;

    d->capabilities = 0;
    portnum = strtol(d->connect, NULL, 0);

    if (portnum == 0) {
        portnum = LPT_PORT;
        info("assuming standard printer port at 0x%x", portnum);
    }

    d->control.hd44780.dataport = portnum;
    d->control.hd44780.controlport = portnum + LPT_CONTROL;

    accessport(d->control.hd44780.dataport);
    accessport(d->control.hd44780.controlport);

    hd44780_setupblock(d);
    hd44780_setupvbar(d);
}
void
hd44780_disconnect(struct driver *d)
{
    bzero(d, sizeof(struct driver));
}
int
hd44780_isconnected(struct driver *d)
{
    /* this driver has no notion of a connection */
    return (d != NULL && d->connect !=NULL &&
            d->control.hd44780.dataport != 0);
}
void
hd44780_clear(struct driver *d)
{
    // TODO
}
char
hd44780_gethchar(struct driver *d, int percentage)
{
    fatal("%s: horizontal bars not implemented", d->driver);

    return 0; /* NOT REACHED */
}
char *
hd44780_getoption()
{
    return "todo";
    // TODO
}
char
hd44780_getvchar(struct driver *d, int percentage)
{
    int unit = 1000 / 8;
    int i;

    percentage %= 101;
    percentage += 5;
    percentage *= 10;

    i = percentage / unit;

    if (i == 0) {
        return HD_CHAR_SPACE;
    }

    if (i == 8) {
        return HD_CHAR_BLOCK;
    }

    return HD_CHAR_VBAR + i - 1;
}
void
hd44780_flush(struct driver *d)
{
    /* Not needed for this driver */
}
void
hd44780_nprintf(struct driver *d, int maxsize, char *fmt, ...)
{
    int i;
    int size;
    va_list args;

    maxsize %= sizeof(d->buflen);

    va_start(args, fmt);
    vsnprintf(d->buffer, maxsize, fmt, args);
    va_end(args);
    size = strlen(d->buffer);

    for (i = 0; i < size; i++) {
        hd44780_writedata(d, d->buffer[i]);
    }
}
void
hd44780_option(struct driver *d, int argc, char **argv)
{
    // TODO
}
void
hd44780_printf(struct driver *d, char *fmt, ...)
{
    int i;
    int size;
    va_list args;

    va_start(args, fmt);
    vsnprintf(d->buffer, d->buflen, fmt, args);
    size = strlen(d->buffer);
    va_end(args);

    for (i = 0; i < size; i++) {
        hd44780_writedata(d, d->buffer[i]);
    }
}
void
hd44780_reset(struct driver *d)
{
    hd44780_writecontrol(d, (HDCMD_FUNCTION |
                              HDCMD_FUNCTION_8BIT |
                              HDCMD_FUNCTION_2LINE |
                              HDCMD_FUNCTION_FONT5X7));
    hd44780_writecontrol(d, (HDCMD_DISPLAY_CONTROL |
                              HDCMD_DISPLAY_CONTROL_DISPLAY_OFF |
                              HDCMD_DISPLAY_CONTROL_CURSOR_OFF |
                              HDCMD_DISPLAY_CONTROL_BLINK_OFF));
    hd44780_writecontrol(d, (HDCMD_DISPLAY_CONTROL |
                              HDCMD_DISPLAY_CONTROL_DISPLAY_ON));
    hd44780_writecontrol(d, (HDCMD_ENTRY_MODE |
                              HDCMD_ENTRY_MODE_CURSOR_INCREMENT));
    hd44780_writecontrol(d, HDCMD_CURSOR_HOME);
}
void
hd44780_setbacklight(struct driver *d, int backlight)
{
    fatal("%s: backlight control not implemented", d->driver);
}
void
hd44780_setcapabilities(struct driver *d, int capabilities)
{
    if (capabilities & DC_BACKLIGHT) {
        warning("%s: backlight support cannot be detected", d->driver);
        capabilities &= ~DC_BACKLIGHT;
    }

    if (capabilities & DC_BLOCK) {
        if (!(d->capabilities & DC_BLOCK)) {
            hd44780_setupblock(d);
        }
        capabilities &= ~DC_BLOCK;
    }

    if (capabilities & DC_CONTRAST) {
        capabilities &= ~DC_CONTRAST;
    }

    if (capabilities & DC_HBAR) {
        hd44780_setuphbar(d);
        capabilities &= ~DC_HBAR;
    }

    if (capabilities & DC_VBAR) {
        hd44780_setupvbar(d);
        capabilities &= ~DC_VBAR;
    }

    if (capabilities != 0) {
        fatal("%s: unknown capability 0x%08x requested", d->driver, capabilities);
    }
}
void
hd44780_setcontrast(struct driver *d, int contrast)
{
    fatal("%s: contrast control not implemented", d->driver);
}
void
hd44780_setspecialchar(struct driver *d, int ch, struct specialchar *data)
{
    int ch_data;
    int i, bit, mask;

    if ((ch < HD_CHAR_SPECIAL) || (ch > HD_CHAR_SPECIAL + 8)) {
        fatal("%s: cannot set specialchar %d",
              d->driver, ch);
    }

    hd44780_writecontrol(d, HDCMD_SETCGRAM | (ch * 8));
    for (i = 0; i < 8; i++) {
        ch_data = 0;
        mask = 0xF;

        for (bit=0; bit<6; bit++) {
            ch_data += ((data->line[i] & mask) ? (1 << bit) : 0);
            mask <<= 4;
        }
        hd44780_writedata(d, ch_data);
    }
}
void
hd44780_setxy(struct driver *d, int x, int y)
{
    d->x = x % d->maxx;
    d->y = y % d->maxy;

    hd44780_writecontrol(d, HDCMD_CURSOR_HOME);
    hd44780_writecontrol(d, HDCMD_SETDDRAM | (hd44780_rowstart[d->y] + d->x));
}
void
hd44780_setupblock(struct driver *d)
{
    struct specialchar block = {
        {
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
        }};

    hd44780_setspecialchar(d, HD_CHAR_BLOCK, &block);
    d->capabilities |= DC_BLOCK;
}
void
hd44780_setuphbar(struct driver *d)
{
    /* only 7 custom chars allowed, we use these for vbars */
    fatal("%s: horizontal bar not supported in this driver", d->driver);
}
void
hd44780_setupvbar(struct driver *d)
{
    int i;
    struct specialchar vbar[] = {
        {{
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x11111,
        }},
        {{
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x11111,
            0x11111,
        }},
        {{
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x11111,
            0x11111,
            0x11111,
        }},
        {{
            0x00000,
            0x00000,
            0x00000,
            0x00000,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
        }},
        {{
            0x00000,
            0x00000,
            0x00000,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
        }},
        {{
            0x00000,
            0x00000,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
        }},
        {{
            0x00000,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
            0x11111,
        }}
    };

    for (i = 0; i < (sizeof(vbar) / sizeof(struct specialchar)); i++) {
        hd44780_setspecialchar(d, HD_CHAR_VBAR + i, &vbar[i]);
    }

    d->capabilities &= !DC_HBAR;
    d->capabilities |= DC_VBAR;
}
void
hd44780_writecontrol(struct driver *d, int data)
{
    writeport(d->control.hd44780.controlport, LPT_CONTROL_MASK);
    writeport(d->control.hd44780.controlport, (HDPIN_ENABLE ^ LPT_CONTROL_MASK));
    waitnsecs(40);
    writeport(d->control.hd44780.dataport, data);
    waitnsecs(40);
    writeport(d->control.hd44780.controlport, LPT_CONTROL_MASK);
}
void
hd44780_writedata(struct driver *d, int data)
{
    writeport(d->control.hd44780.controlport, (HDPIN_RS ^ LPT_CONTROL_MASK));
    writeport(d->control.hd44780.controlport, ((HDPIN_ENABLE | HDPIN_RS) ^ LPT_CONTROL_MASK));
    waitnsecs(40);
    writeport(d->control.hd44780.dataport, data);
    waitnsecs(40);
    writeport(d->control.hd44780.controlport, LPT_CONTROL_MASK);
}
