/*
 * Copyright (c) 2007 Willem Dijkstra
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <termios.h>

#include "drivers.h"
#include "driver_cf.h"
#include "driver_cfp.h"
#include "driver_hd44780.h"
#include "driver_test.h"
#include "error.h"
#include "xmalloc.h"

struct model models[] = {
    { "cf632",     "cf",      16, 2, B19200, {CF_SPECIAL, CF_SPECIAL_END, CF_SPECIAL_WIDTH, CF_SPECIAL_HEIGTH}}, /* assume v2 */
    { "cf632v1",   "cf",      16, 2, B9600,  {CF_SPECIAL, CF_SPECIAL_END, CF_SPECIAL_WIDTH, CF_SPECIAL_HEIGTH}},
    { "cf632v2",   "cf",      16, 2, B19200, {CF_SPECIAL, CF_SPECIAL_END, CF_SPECIAL_WIDTH, CF_SPECIAL_HEIGTH}},
    { "cf634",     "cf",      20, 4, B19200, {CF_SPECIAL, CF_SPECIAL_END, CF_SPECIAL_WIDTH, CF_SPECIAL_HEIGTH}}, /* assume v2 */
    { "cf634v1",   "cf",      20, 4, B9600,  {CF_SPECIAL, CF_SPECIAL_END, CF_SPECIAL_WIDTH, CF_SPECIAL_HEIGTH}},
    { "cf634v2",   "cf",      20, 4, B19200, {CF_SPECIAL, CF_SPECIAL_END, CF_SPECIAL_WIDTH, CF_SPECIAL_HEIGTH}},
    { "cf635",     "cfp",     20, 4, B115200, {CFP_SPECIAL, CFP_SPECIAL_END, CFP_SPECIAL_WIDTH, CFP_SPECIAL_HEIGTH}}, /* assume v1 */
    { "cf635v1",   "cfp",     20, 4, B115200, {CFP_SPECIAL, CFP_SPECIAL_END, CFP_SPECIAL_WIDTH, CFP_SPECIAL_HEIGTH}},
    { "hd44780",   "hd44780", 16, 2, 0, {HD_SPECIAL, HD_SPECIAL_END, HD_SPECIAL_WIDTH, HD_SPECIAL_HEIGTH}}, /* assume y = 2 */
    { "hd44780y2", "hd44780", 16, 2, 0, {HD_SPECIAL, HD_SPECIAL_END, HD_SPECIAL_WIDTH, HD_SPECIAL_HEIGTH}},
    { "hd44780y4", "hd44780", 16, 4, 0, {HD_SPECIAL, HD_SPECIAL_END, HD_SPECIAL_WIDTH, HD_SPECIAL_HEIGTH}},
    { "test",      "test",    16, 4, 0, {0, 0, 0, 0}},
    { NULL,        NULL,       0, 0, 0, {0, 0, 0, 0}}
};

struct driverfmap drivers[] =  {
    { "cf", cf_connect, cf_disconnect, cf_isconnected,
          cf_clear, cf_gethchar, cf_getvchar, cf_flush,
          cf_nprintf, cf_printf, cf_reset, cf_setbacklight,
          cf_setcapabilities, cf_setcontrast, cf_setspecialchar,
          cf_setupblock, cf_setuphbar, cf_setupvbar,
          cf_setxy, cf_getoption, cf_option},
    { "cfp", cfp_connect, cfp_disconnect, cfp_isconnected,
          cfp_clear, cfp_gethchar, cfp_getvchar, cfp_flush,
          cfp_nprintf, cfp_printf, cfp_reset, cfp_setbacklight,
          cfp_setcapabilities, cfp_setcontrast, cfp_setspecialchar,
          cfp_setupblock, cfp_setuphbar, cfp_setupvbar,
          cfp_setxy, cfp_getoption, cfp_option},
    {"hd44780", hd44780_connect, hd44780_disconnect, hd44780_isconnected,
          hd44780_clear, hd44780_gethchar, hd44780_getvchar, hd44780_flush,
          hd44780_nprintf, hd44780_printf, hd44780_reset, hd44780_setbacklight,
          hd44780_setcapabilities, hd44780_setcontrast, hd44780_setspecialchar,
          hd44780_setupblock, hd44780_setuphbar, hd44780_setupvbar,
          hd44780_setxy, hd44780_getoption, hd44780_option},
    { "test", test_connect, test_disconnect, test_isconnected,
          test_clear, test_gethchar, test_getvchar, test_flush,
          test_nprintf, test_printf, test_reset, test_setbacklight,
          test_setcapabilities, test_setcontrast, test_setspecialchar,
          test_setupblock, test_setuphbar, test_setupvbar,
          test_setxy, test_getoption, test_option},
    { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
          NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL}
};

struct driverfmap *
finddriverfmap(char *drivername)
{
    int i;

    for (i = 0; drivers[i].name != NULL; i++) {
        if (strncmp(drivers[i].name, drivername, strlen(drivername)) == 0) {
            return &drivers[i];
        }
    }

    return NULL;
}

struct model *
findmodel(char *modelname)
{
    int i;

    if (modelname != NULL) {
        for (i = 0; models[i].name != NULL; i++) {
            if (strncmp(models[i].name, modelname, strlen(models[i].name)) == 0) {
                return &models[i];
            }
        }
    }

    return NULL;
}

struct driver *
getdriver(char *modelname, char *connectstring)
{
    struct driver *d;
    struct model *m = NULL;

    d = xmalloc(sizeof(struct driver));
    bzero(d, sizeof(struct driver));

    if (connectstring != NULL) {
        d->connect = xstrdup(connectstring);
    }

    if ((m = findmodel(modelname)) != NULL) {
        d->driver = m->driver;
        d->maxx = m->maxx;
        d->maxy = m->maxy;
        d->model = xstrdup(m->name);
        d->speed = m->speed;
    } else {
        fatal("could not find model '.200s'", modelname);
    }

    d->buflen = d->maxx * d->maxy * 2;
    d->buffer = xmalloc(d->buflen);

    if ((d->fmap = finddriverfmap(m->driver)) == NULL) {
        fatal("could not find driver for model '%.200s'", modelname);
    }

    return d;
}
int
getmodelinfo(char *modelname, char *buffer, size_t maxsize)
{
    struct model *m = findmodel(modelname);

    if (m != NULL) {
        return snprintf(
            buffer, maxsize,
            "%12s %12s % 4d % 4d % 8d % 6d % 6d % 6d % 7d",
            m->name,
            m->driver,
            m->speed,
            m->maxx,
            m->maxy,
            m->special.start,
            m->special.stop,
            m->special.height,
            m->special.width);
    }

    return 0;
}
int
getmodelinfoheader(char *buffer, size_t maxsize)
{
    return snprintf(
        buffer, maxsize,
        "%12s %12s %4s %4s %8s %6s %6s %6s %7s",
        "model",
        "driver",
        "maxx",
        "maxy",
        "speed",
        "sstart",
        "send",
        "swidth",
        "sheight");
}
int
getmodels(char *buffer, size_t maxsize)
{
    int i;
    int printed = 0;

    if (maxsize < 1) {
        return 0;
    }

    for (i = 0; models[i].name != NULL && ((maxsize - 1 - printed) > 0); i++) {
        printed += snprintf(buffer + printed, (maxsize - 1 - printed), "%s ", models[i].name);
    }
    buffer[maxsize - 1] = '\0';

    return i;
}
