/*
 * Copyright (c) 2003-2007 Willem Dijkstra
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "error.h"
#include "lcd.h"
#include "xmalloc.h"

void
lcd_clear(struct lcd *l)
{
    if (l != NULL && l->fb != NULL) {
	memset(l->fb, ' ', l->d->maxx * l->d->maxy);
    }
}
struct lcd *
lcd_connect(struct driver *d)
{
    struct lcd *l;

    if (d == NULL) {
	return NULL;
    }

    l = xmalloc(sizeof(struct lcd));
    l->d = d;
    l->fb = xmalloc(d->maxx * d->maxy);

    lcd_clear(l);

    return l;
}
void
lcd_flush(struct lcd *l)
{
    int i;
    struct driver *d = l->d;
    struct driverfmap *f = l->d->fmap;

    for (i = 0; i < l->d->maxy; i++) {
	(f->setxy)(d, 0, i);
	(f->nprintf)(d, d->maxx + 1, l->fb + (i * d->maxx));
    }
}
void
lcd_hbar(struct lcd *l, int x, int y, int mx, int percentage)
{
    fatal("lcd_hbar is not implemented");
}
void
lcd_printf(struct lcd *l, int x, int y, char *fmt, ...)
{
    va_list args;
    char buf[LCD_MAXSTR];
    int size, i;

    va_start(args, fmt);
    vsnprintf(buf, sizeof(buf), fmt, args);
    va_end(args);
    size = strlen(buf);

    if (size > LCD_MAXSTR || size < 0) {
	size = LCD_MAXSTR;
    }

    if ((x < 0) || (x >= l->d->maxx) ||
	(y < 0) || (y >= l->d->maxy) ||
	(x + size) > l->d->maxx ) {
	warning("lcd_printf(%d,%d,'%.32s') breaks lcd screen boundaries",
		x, y, &buf[0]);
    }

    for (i=0; i<size; i++) {
	lcd_set(l, x+i, y, buf[i]);
    }
}
void
lcd_reset(struct lcd *l)
{
    lcd_clear(l);
    (l->d->fmap->reset)(l->d);
}
void
lcd_set(struct lcd *l, int x, int y, char c)
{
    if ((x < 0) || (x >= l->d->maxx) ||
	(y < 0) || (y >= l->d->maxy))
	warning("lcd_set(%d,%d,'%c') breaks lcd screen boundaries",
	      x, y, c);

    x %= l->d->maxx + 1;
    y %= l->d->maxy + 1;
    l->fb[x + (y * l->d->maxx)] = c;
}
void
lcd_vbar(struct lcd *l, int x, int y, int my, int percentage)
{
    char c;
    int unit;
    int cell;
    struct driverfmap *f = l->d->fmap;

    unit = 100 / ((y - my) + 1);

    for (cell = y; cell >= my && percentage >= 0; cell--) {
	if (percentage > unit) {
	    c = (f->getvchar)(l->d, 100);
	} else {
	    c = (f->getvchar)(l->d, percentage);
	}
	lcd_set(l, x, cell, c);
	percentage -= unit;
    }
}
void
lcd_printunit(struct lcd *l, int x, int y, int msp, int lsp, double value)
{
    char buf[LCD_MAXSTR];
    char *units = " kMGTPEZY";
    int i;

    snprintf(&buf[0], sizeof(buf), "%.0f", value);
    for (i = 0; i < strlen(units) && strlen(buf) > msp; i++) {
	value /= 1000;
	snprintf(&buf[0], sizeof(buf), "%.0f", value);
    }

    snprintf(&buf[0], sizeof(buf), "%%%d.%df%c", msp, lsp, units[i]);
    lcd_printf(l, x, y, buf, value);
}
