/*
 * Copyright (c) 2005-2007 Willem Dijkstra
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <limits.h>
#include <sysexits.h>
#include <string.h>
#include <unistd.h>

#include "drivers.h"
#include "error.h"
#include "sylcd.h"
#include "xmalloc.h"

__BEGIN_DECLS
void modelinfo(char *lcdmodel);
void usage();
__END_DECLS

int buflen = _POSIX2_LINE_MAX;
char buffer[_POSIX2_LINE_MAX];
/* Decode <special char number>:<row 1 bits>:<row 2 bits>:...:<row 8 bits>
 * bits are written as 1 or 0, bitmasks are considered to be lsb
 */
int
decode_specialchar(char *customdef, struct specialchar *data)
{
    char *word, *rest;
    int row = 0;
    int i;
    int ch;

    rest = customdef;
    word = strsep(&rest, ":");

    if (word == NULL || rest == NULL) {
	fatal("expected special string of the form <char number>:<bitrow 1>:<bitrow 2>:...");
    }

    ch = strtol(word, NULL, 0);

    while (rest != NULL && row < 8) {
	word = strsep(&rest, ":");
	data->line[row] = 0;
	for (i = 0; i < strlen(word); i++) {
	    data->line[row] <<= 4;
	    data->line[row] |= ((word[i] == '0') ? 0 : 1);
	}
	row++;
    }

    return ch;
}
int
main(int argc, char *argv[])
{
    struct specialchar customchar;
    struct driver *d;
    char *customdef, *lcdmodel, *lcdport;
    char specialch;
    int ch, i, x, y;
    int flag_flush, flag_info, flag_reset, flag_special;

    customdef = NULL;
    lcdmodel = NULL;
    lcdport = NULL;
    flag_debug = 0;
    flag_info = 0;
    flag_flush = 0;
    flag_reset = 0;
    flag_special = -1;

    x = -1;
    y = -1;

    while ((ch = getopt(argc, argv, "c:dfi:m:p:rs:vx:y:")) != -1) {
	switch (ch) {
	case 'c':
	    customdef = xstrdup(optarg);
	    break;

	case 'd':
	    flag_debug += 1;
	    break;

	case 'f':
	    flag_flush = 1;
	    break;

	case 'i':
            flag_info = 1;
	    if (optarg) {
		lcdmodel = xstrdup(optarg);
	    } else {
                lcdmodel = xstrdup("all");
	    }
	    break;

	case 'm':
	    if (optarg) {
		lcdmodel = xstrdup(optarg);
	    } else {
		warning("Need a <lcd model>");
		usage();
	    }
	    break;

	case 'p':
	    if (optarg) {
		lcdport = xstrdup(optarg);
	    } else {
		warning("Need a <lcd port>");
		usage();
	    }
	    break;

	case 'r':
	    flag_reset = 1;
	    break;

	case 's':
	    flag_special = 1;
	    specialch = strtol(optarg, NULL, 0);
	    break;

	case 'x':
	    x = strtol(optarg, NULL, 0);
	    break;

	case 'y':
	    y = strtol(optarg, NULL, 0);
	    break;

	case 'v':
	    info("lcdprint version " SYLCD_VERSION);
	    /* drop through */

	default:
            usage();
	}
    }

    if (flag_info) {
        modelinfo(lcdmodel);
        return 0;
    }

    if (lcdmodel == NULL || lcdport == NULL) {
        fatal("need a lcdmodel and a lcdport");
    }

    d = getdriver(lcdmodel, lcdport);
    if (d == NULL) {
        fatal("could not connect to '%.200s' @ '%.200s'", lcdmodel, lcdport);
    }

    (*d->fmap->connect)(d);

    if (customdef != NULL) {
	ch = decode_specialchar(customdef, &customchar);
        (*d->fmap->setspecialchar)(d, ch, &customchar);
    }

    if (flag_reset) {
        (*d->fmap->reset)(d);
    }

    if (x != -1 && y != -1) {
        (*d->fmap->setxy)(d, x, y);
    } else if (x != -1 || y != -1) {
	fatal("set both x and y values when positioning");
    }

    if (flag_special != -1) {
        (*d->fmap->nprintf)(d, 2, &specialch);
    }

    argc -= optind;
    argv += optind;

    for (i = 0; i < argc; i++) {
        (*d->fmap->printf)(d, argv[i]);
    }

    if (flag_flush != -1) {
        (*d->fmap->flush)(d);
	sleep(1);
    }

    return 0;
}
void modelinfo(char *lcdmodel)
{
    if (getmodelinfo(lcdmodel, &buffer[0], buflen) > 0) {
        if (getmodelinfoheader(&buffer[0], buflen) > 0) {
            info(buffer);
        }
        getmodelinfo(lcdmodel, &buffer[0], buflen);
        info(buffer);
    } else {
        if (getmodels(&buffer[0], buflen) > 0) {
            fatal("unknown lcd model, compiled in models: %s", &buffer[0]);
        } else {
            fatal("unknown lcd model and unable to retrieve compiled in models");
        }
    }
}

void usage() {
    info("usage: %s [-c num:bitrow1:..] [-d] [-f] [-i <lcd model>]", __progname);
    info(" <-m lcd model> <-p lcd port>");
    info(" [-r] [-s num] [-t <text>] [-x <x> -y <y>] [<text>]\n");
    info(" -c num:bitrow1:...  define special character num");
    info("    num = special height x special width rows of bits");
    info("    e.g. -c 128:11111:00000:11111:00000:11111:00000:11111");
    info(" -d                  enable debug information");
    info(" -f                  flush after print; needed for non-blocking lcd devices");
    info(" -i <lcd model>      show model info, includes special chars and their sizes");
    info(" -m <lcd model>      mandatory; select lcd model");
    info(" -p <lcd port>       mandatory; driver dependent lcd connection");
    info("    e.g -p /dev/ttyU0 for the serial drivers");
    info("        -p 0x378 for the parallel port drivers");
    info(" -r                  reset lcd");
    info(" -s num              print special character num");
    info(" -v                  version information");
    info(" -x <x> -y <y>       move lcd cursor to x, y");

    if (getmodels(&buffer[0], buflen) > 0) {
        info("\n supported lcd models: %s", &buffer[0]);
    }

    exit(EX_USAGE);
}
