/*
 * Copyright (c) 2003-2005 Willem Dijkstra
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>

#include "error.h"
#include "net.h"

void
init_muxconn(struct muxconn *m, char *host, char *service)
{
    struct addrinfo hints;
    int error;

    bzero((void *) &hints, sizeof(struct addrinfo));
    bzero((void *) m, sizeof(struct muxconn));

    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    debug("mux host = %s %s", host, service);

    if ((error = getaddrinfo(host, service, &hints, &m->resolved)) != 0)
	fatal("could not get address information for %s:%s - %s",
	      host, service, gai_strerror(error));

    connect_muxconn(m);
}

void
connect_muxconn(struct muxconn *m)
{
    static int warn = 0;
    uint32_t sock;

    sock = rbuffer_fd(&m->rb);
    rbuffer_clear(&m->rb);
    close(sock);

    debug("connecting to muxhost");

    if ((sock = socket(m->resolved->ai_family, SOCK_STREAM, 0)) == -1)
	fatal("could not obtain socket: %.200s", strerror(errno));

    if (connect(sock, (struct sockaddr *) m->resolved->ai_addr, m->resolved->ai_addrlen) == -1) {
	if (warn == 0) {
	    warning("could not bind socket: %.200s", strerror(errno));
	    warn++;
	}
	sleep(30);
    } else {
	warn = 0;
    }

    fcntl(sock, F_SETFL, O_NONBLOCK);
    rbuffer_attach(&m->rb, sock);
}

size_t
get_muxdata(struct muxconn *m)
{
    size_t seen;

    if (rbuffer_get_data(&m->rb, &seen) == 0) {
	if (errno != 0 && errno != EAGAIN) {
	    debug("errno = %d; %s", errno, strerror(errno));
	    connect_muxconn(m);
	    return (size_t) 0;
	}
    }

    return seen;
}

int
get_line(struct muxconn *m, char **s, size_t *maxlen)
{
    return rbuffer_get_line(&m->rb, s, maxlen);
}
