/*
 * Copyright (c) 2007 Willem Dijkstra
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/select.h>

#include <string.h>
#include <unistd.h>

#include "error.h"
#include "rbuffer.h"

void
rbuffer_attach(struct rbuffer *rb, uint32_t fd)
{
    rb->fd = fd;
}

void
rbuffer_clear(struct rbuffer *rb)
{
    bzero(rb, sizeof(struct rbuffer));
}

void
rbuffer_compact(struct rbuffer *rb)
{
    if (rb->start != 0) {
	if (rb->start < rb->end) {
	    /* there is some data left, move it to the start of the buffer */
	    bcopy(&rb->data[rb->start], &rb->data, rb->end - rb->start);
	    rb->end  -= rb->start;
	    rb->start = 0;
	    bzero(&rb->data[rb->end], (sizeof(rb->data) - rb->end));
	} else {
	    rbuffer_clear(rb);
	}
    }
}

uint32_t
rbuffer_fd(struct rbuffer *rb)
{
    return rb->fd;
}

uint32_t
rbuffer_get_byte(struct rbuffer *rb, uint8_t *byte)
{
    if (rbuffer_len(rb) > 0) {
	*byte = rb->data[rb->start];
	rb->start++;
	return 1;
    } else {
	return 0;
    }
}

uint32_t
rbuffer_get_data(struct rbuffer *rb, size_t *len)
{
    if (rb->start > sizeof(struct rbuffer) / 2)
	rbuffer_compact(rb);

    if (rb->end >= sizeof(rb->data)) {
	warning("rbuffer is full");
	rbuffer_clear(rb);
    }

    *len = read(rb->fd,
                (void *) &rb->data[rb->end],
                (size_t) (sizeof(rb->data) - rb->end));

    if (*len == -1 || *len == 0) {
	return 0;
    }

    rb->end += *len;

    return 1;
}

uint32_t
rbuffer_get_data_wait(struct rbuffer *rb, size_t *len, struct timeval *wait)
{
    int socksactive;
    fd_set readset;

    FD_ZERO(&readset);
    FD_SET(rb->fd, &readset);

    socksactive = select(rb->fd + 1, &readset, NULL, NULL, wait);

    if (socksactive != 1) {
	return 0;
    }

    return rbuffer_get_data(rb, len);
}

uint32_t
rbuffer_get_line(struct rbuffer *rb, char **s, size_t *maxlen)
{
    uint32_t i;

    if (rb->start > sizeof(struct rbuffer) / 2)
	rbuffer_compact(rb);

    for (i = rb->start; i < rb->end && rb->data[i] != '\n'; i++);

    if (rb->data[i] != '\n') {
	if (rb->end >= sizeof(rb->data)) {
	    warning("symux data parse error; no linefeed even after %d bytes",
		    rb->end);
	    rbuffer_clear(rb);
	}
	return 0;
    }

    rb->data[i] = 0;
    *s = &rb->data[rb->start];
    *maxlen = (rb->end - rb->start);
    rb->start = i + 1;

    return 1;
}

uint32_t
rbuffer_len(struct rbuffer *rb)
{
    return rb->end - rb->start;
}

uint32_t
rbuffer_unget_byte(struct rbuffer *rb, uint8_t byte)
{
    if (rb->start == 0) {
        return 0;
    } else {
        rb->start--;
        rb->data[rb->start] = byte;
        return 1;
    }
}
