/*
 * Copyright (c) 2003-2007 Willem Dijkstra
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>

#include "error.h"
#include "lcd.h"
#include "st_if.h"
#include "stream.h"
#include "xmalloc.h"

void
if_init(struct lcd *l, struct stream *st)
{
    if (st->data != NULL) {
	warning("%s:%d:st->data != NULL", __FILE__, __LINE__);
	xfree(st->data);
    }

    st->data = xmalloc(sizeof(struct ps_if) * l->d->maxx);
    bzero(st->data, sizeof(struct ps_if) * l->d->maxx);
}

void
if_read(struct lcd *l, struct stream *st, char *s, size_t maxlen)
{
    struct ps_if  if_data;
    struct ps_if  *last, *curr;
    int data_index, i;
    long unsigned int timestamp, dt, dbytes_in, dbytes_out;

    timestamp = 0;
    i = (st->index + 1) % l->d->maxx;

    /* jump over if:arg: */
    data_index = get_nextarg_index(s, maxlen, 2);

    if (1 != sscanf(&s[data_index], "%lu:", &timestamp))
	return;

    /* jump over timestamp : packets in: packets out */
    data_index += get_nextarg_index(&s[data_index], maxlen - data_index, 3);

    if (2 != sscanf(&s[data_index], "%lu:%lu:",
		    &if_data.bytes_in, &if_data.bytes_out))
	return;

    /* all data was read succesfully - commit */
    curr = &((struct ps_if *)st->data)[i];
    last = &((struct ps_if *)st->data)[st->index];

    curr->bytes_in   = if_data.bytes_in;
    curr->bytes_out  = if_data.bytes_out;
    curr->timestamp  = timestamp;
    dt = timestamp - last->timestamp;

    /* can only compute difference after 2nd read */
    if (st->lastupdate != 0) {
	dbytes_in        = if_data.bytes_in - last->bytes_in;
	dbytes_out       = if_data.bytes_out - last->bytes_out;

	/* handle overflows */
	if (last->bytes_in > if_data.bytes_in)
	    dbytes_in = 0;
	if (last->bytes_out > if_data.bytes_out)
	    dbytes_out = 0;

	curr->bps_in     = (dbytes_in << 3) / dt;
	curr->bps_out    = (dbytes_out << 3) / dt;

    }

    st->index = i;
    st->lastupdate = timestamp;
}
/* determine maxima over last <positions> */
void
if_get_max(struct stream *st, int maxx, double *max_in, double *max_out, int positions)
{
    int c, index;
    struct ps_if *curr;

    *max_in = 0.0;
    *max_out = 0.0;

    if (positions > maxx)
	fatal("Requested more positions precision than stored");

    if (positions <= 0)
	positions = maxx;

    for (c = 0; c < positions; c++) {
	if ((index = (st->index - c)) < 0) {
	    index += maxx;
	}
	curr = &((struct ps_if *)st->data)[index];

	if (curr->bps_in > *max_in)
	    *max_in      = curr->bps_in;

	if (curr->bps_out > *max_out)
	    *max_out     = curr->bps_out;
    }
}
/* Mode IFMODE_INOUT
 *
 * +0123456789012345+
 * 0ifn: 123Ki 123Ko0
 * 1-=-=-=-i-=-=-=-o1
 * +0123456789012345+
 */
void
if_show(struct lcd *l, struct stream *st, int mode, int startx, int starty, int stopx, int stopy)
{
    double max_din, max_dout, max_dinout, percent;
    struct ps_if *curr;
    int i, ri, bar, half;

    max_din = 0.0;
    max_dout = 0.0;

    switch (mode) {
    case IFMODE_INOUT:
	half = (stopx - startx) / 2;
	if_get_max(st, l->d->maxx, &max_din, &max_dout, half);
	max_dinout = (max_dout > max_din) ? max_dout : max_din;
	percent = (max_dinout / 100);
	if (percent == 0) {
	    percent = 1;
	}

	for (i = 0; i < half - 1; i++) {
	    if ((ri = (st->index - i)) < 0) {
		ri += (stopx - startx);
	    }
	    curr = &((struct ps_if *)st->data)[ri];

	    bar = (int)(curr->bps_in / percent);
	    lcd_vbar(l, half  - 2 - i, stopy - 1, starty + 1, bar);

	    bar = (int)(curr->bps_out / percent);
	    lcd_vbar(l, stopx - 2 - i, stopy - 1, starty + 1, bar);
	}

	lcd_printf(l, half  - 1, starty + 1, "%c", 'i');
	lcd_printf(l, stopx - 1, starty + 1, "%c", 'o');
	break;

    default:
	fatal("illegal interface display mode selected");
    }

    curr = &((struct ps_if *)st->data)[st->index];

    lcd_printf(l, startx, starty, "%4s", st->sargs);
    lcd_printunit(l, startx + 5, starty, 3, 0, curr->bps_in);
    lcd_printf(l, startx + 9, starty, "%c", 'i');
    lcd_printunit(l, stopx - 5, starty, 3, 0, curr->bps_out);
    lcd_printf(l, stopx - 1, starty, "%c", 'o');
}
