/*
 * Copyright (c) 2003-2007 Willem Dijkstra
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>

#include "drivers.h"
#include "error.h"
#include "lcd.h"
#include "net.h"
#include "rbuffer.h"
#include "st_if.h"
#include "sylcd.h"
#include "stream.h"
#include "xmalloc.h"

__BEGIN_DECLS
void usage();
__END_DECLS

int
init_streams(struct lcd *l, struct streamlist *sl, char *stream_host, char *stream_ignore, char *s, size_t maxlen)
{
    int i, items = 0;
    struct stream *st;
    char *end;
    char buffer[_POSIX2_LINE_MAX];

    while ((i = get_stream_index(s, maxlen, "if", NULL)) != -1) {

	/* jump over 'if:' */
	i += 3;
	maxlen -= i;
	s += i;

	/* seek end of if definition 'xl0:' */
	if ((end = memchr(s, ':', maxlen)) == NULL)
	    return items;

	*end = '\0';
        /* determine if this interface is not ignored */
        snprintf(buffer, _POSIX2_LINE_MAX, "if(%.200s)", s);
        buffer[_POSIX2_LINE_MAX - 1] = '\0';
        if (strstr(stream_ignore, &buffer[0]) == NULL) {
            if (find_stream(sl, stream_host, "if", s) == NULL) {
                if ((st = create_stream(stream_host, "if", s)) == NULL)
                    fatal("%s:%d:could not create stream", __FILE__, __LINE__);

                if_init(l, st);
                debug("found interface %s", s);
                items++;

                SLIST_INSERT_HEAD(sl, st, next);
            }
        }
        /* jump over end of string and rerun */
	i = (end - s) + 1;
	maxlen -= i;
        s += i;
    }

    return items;
}

int
main(int argc, char** argv)
{
    struct muxconn m;
    struct streamlist sl;
    struct lcd *l;
    struct driver *d;
    struct stream *st, *show_st;
    char *streamignore, *lcdmodel, *lcdport, *muxhost, *muxport, *streamhost, *s;
    int i, sleeper, items;
    char ch;
    size_t maxlen;

    flag_debug = 0;
    flag_daemon = 0;
    streamhost = DEFAULT_SOURCE_HOST;
    muxhost = DEFAULT_MUX_HOST;
    muxport = DEFAULT_MUX_PORT;
    lcdport = NULL;
    lcdmodel = NULL;
    sleeper = 0;
    st = show_st = NULL;
    SLIST_INIT(&sl);

    while ((ch = getopt(argc, argv, "di:H:m:p:v")) != -1) {
	switch (ch) {
	case 'd':
	    flag_debug += 1;
	    break;

        case 'i':
            if (optarg) {
                streamignore = xstrdup(optarg);
            } else {
                warning("Need a commaseperated string of streams to ignore");
            }
            break;

	case 'H':
	    if (optarg) {
		streamhost = xstrdup(optarg);
	    } else {
		warning("Need a <host to display>");
		usage();
	    }
	    break;

	case 'm':
	    if (optarg) {
		lcdmodel = xstrdup(optarg);
	    } else {
		warning("Need a <lcd model>");
		usage();
	    }
	    break;

	case 'p':
	    if (optarg) {
		lcdport = xstrdup(optarg);
	    } else {
		warning("Need a <lcd port>");
		usage();
	    }
	    break;

	case 'v':
	    info("sylcd version " SYLCD_VERSION);
	    /* drop through */
	default:
	    usage();
	    break;
	}
    }

    argc -= optind;
    argv += optind;

    switch (argc) {
    case 2:
	muxhost = xstrdup(*argv);
	argv++;
	muxport = xstrdup(*argv);
	break;
    case 1:
	muxhost = xstrdup(*argv);
	break;
    case 0:
	break;
    default:
	fatal("Too many arguments");
    }

    setegid(getgid());
    setgid(getgid());

#if !defined TEST
    if (flag_debug == 0) {
	if (daemon(0, 0) != 0)
	    fatal("daemonize failed");

	flag_daemon = 1;
    }
#endif

    d = getdriver(lcdmodel, lcdport);

    if ((d == NULL) || (d->fmap == NULL)) {
        fatal("driver init failed");
    }

    (d->fmap->connect(d));

    l = lcd_connect(d);
    lcd_reset(l);
    lcd_printf(l, 0, 0, "sylcd " SYLCD_VERSION);
    lcd_flush(l);

    bzero(&m, sizeof(m));
    init_muxconn(&m, muxhost, muxport);

    debug("detecting interfaces in %s host streams", streamhost);

    lcd_clear(l);
    i = 0;
    /* wait n seconds or m bytes, whichever is first */
    while ((i++ < SYLCD_INIT_TIME) &&
	   (rbuffer_len(&m.rb) < SYLCD_INIT_SIZE)) {
	get_muxdata(&m);
	lcd_printf(l, 0, 0, "waiting for data");
	lcd_printunit(l, 0, 1, 3, 0, rbuffer_len(&m.rb));
	lcd_printf(l, 4, 1, "b received");
	lcd_flush(l);
	sleep(1);
    }

    while (get_line(&m, &s, &maxlen)) {
	if (!is_host_stream(s, maxlen, streamhost))
	    continue;

	items = init_streams(l, &sl, streamhost, streamignore, s, maxlen);
    }

    if (SLIST_EMPTY(&sl))
	fatal("no interfaces found for %s in streams from %s %s",
	      streamhost, muxhost, muxport);

    info("sylcd " SYLCD_VERSION);
    show_st = SLIST_FIRST(&sl);

    for (;;) {
	if (get_muxdata(&m)) {
	    while (get_line(&m, &s, &maxlen)) {

		if (!is_host_stream(s, maxlen, streamhost))
		    break;

		SLIST_FOREACH(st, &sl, next)
		    if ((i = get_stream_index(s, maxlen, st->stype, st->sargs)) > 0)
			if_read(l, st, &s[i], maxlen - i);
	    }

	    /* TODO: rewrite display code */
	    if (d->maxy == 2 || flag_debug == 2)
	    {
		lcd_clear(l);
		if_show(l, show_st, IFMODE_INOUT, 0, 0, d->maxx, d->maxy);
		lcd_flush(l);
	    } else if (d->maxy == 4) {
		lcd_clear(l);
		if_show(l, show_st, IFMODE_INOUT, 0, 0, d->maxx, d->maxy / 2);
		if (SLIST_NEXT(show_st, next) != NULL) {
		    if_show(l, SLIST_NEXT(show_st, next), IFMODE_INOUT, 0, (d->maxy / 2), d->maxx, d->maxy);
		} else {
		    if_show(l, SLIST_FIRST(&sl), IFMODE_INOUT, 0, (d->maxy / 2), d->maxx, d->maxy);
		}
		lcd_flush(l);
	    }

	    if (items > 2) {
                sleeper++;
                if (sleeper > 3) {
                    if (show_st == NULL ||
                        SLIST_NEXT(show_st, next) == SLIST_END(show_st)) {
                        show_st = SLIST_FIRST(&sl);
                    } else {
                        show_st = SLIST_NEXT(show_st, next);
                    }
                    sleeper = 0;
                }
	    }
	}
	sleep(1);
    }
}
void usage() {
    info("usage: %s \t[-d] [-m <lcd model>] [-v] [-H <host to display>]\n"
	 "\t\t[-c <lcd connection>] [<symux host>] [<symux port>]\n", __progname);
    info("\t-d                     debug, also prevents becoming a daemon");
    info("\t-v                     version information");
    info("\t-H <host to display>   which host data should be displayed");
    info("\t                       default: " DEFAULT_SOURCE_HOST);
    info("\t-c <lcd connection>    driver dependent lcd connection");
    info("\t-m <lcd model>         lcd model:");
    
    info("\t<symux host>           symux to connect to");
    info("\t                       default: " DEFAULT_MUX_HOST);
    info("\t<symux port>           symux port to connect to");
    info("\t                       default: " DEFAULT_MUX_PORT);

    exit(EX_USAGE);
}
